clear; format long; addpath('file_library'); close all
global b k a

%% SETUP-------------------------------------------------------------------
elementType = 1;             % 1: spherical triangles; 2: angular elements
maxNumberOfElements = 4000;
integrationOrder = 1;        %  0: Burkardt algorithm; 1-4: Gaussian quadrature rule of order 1 to 4
rhoMax = 0.001;              %  0: uniform refinement; >0: max. value of rho_n per element
b = -1;   % positive: symmetric von Mises distribution (see eq.(3.6) in Li et al. [J R Soc Interface], 2018)
          % negative: symmetric von Mises distribution (see eq.(40) in Rolf-Pissarczyk et al. [Comput Methods Appl Mech Eng], 2021) 
          %           or out-of-plane concentration parameter, non-symmetric von Mises distribution (see eq.(3.6))
a = 10;   % positive: in-plane concentration parameter; non-symmetric von Mises distribution (see eq.(3.6))
          %        0: symmetric von Mises distribution
k = 2*sqrt(2*b/pi)/(besseli(0,a)*erfi(sqrt(2*b)));  % factor used used in all fibre distribution functions

% create initial mesh with four elements:
if elementType == 1 % spherical triangles
    v1 = [1 0 0]; v2 = [0 1 0]; v3 = [0 0 1]; v4 = [-1 0 0]; v5 = [0 0 -1];
    connectivity = zeros(3,3,4);
    connectivity(1,:,1)=v3; connectivity(2,:,1)=v1; connectivity(3,:,1)=v2;
    connectivity(1,:,2)=v1; connectivity(2,:,2)=v5; connectivity(3,:,2)=v2;
    connectivity(1,:,3)=v5; connectivity(2,:,3)=v4; connectivity(3,:,3)=v2;
    connectivity(1,:,4)=v4; connectivity(2,:,4)=v3; connectivity(3,:,4)=v2;
    elementsForRefinement = 1:size(connectivity,3);
    nOfElements = size(connectivity,3);
    
elseif elementType == 2 % angular elements
    connectivity = [0 pi/2 0 pi/2;
        pi/2 pi 0 pi/2;
        pi/2 pi pi/2 pi;
        0 pi/2 pi/2 pi];
    elementsForRefinement = 1:1:size(connectivity,1);
    nOfElements = size(connectivity,1);
end


%% NUMERICAL INTEGRATION---------------------------------------------------
c = 0; % initialise refinement level counter
while isempty(elementsForRefinement)~=1
    c = c + 1;
    
    % refine marked surface elements
    for i=1:length(elementsForRefinement)
        iElem = elementsForRefinement(i);
        [connectivity,nOfElements] = localRefinement(connectivity,iElem,elementType);
    end
    
    elementsForRefinement = []; rho = [];
    
    % integrate fibre distribution
    for iElem=1:nOfElements
        
        if elementType == 1 % spherical triangles
            I = sphericalTriangleIntegration(connectivity,iElem,integrationOrder);
        
        elseif elementType == 2 % angular elements
            theta1=connectivity(iElem,1);theta2=connectivity(iElem,2);
            phi1=connectivity(iElem,3);phi2=connectivity(iElem,4);
            I = sphericalElementIntegration(theta1,theta2,phi1,phi2,integrationOrder);
        end
        
        rho = [rho;I/(2*pi)]; % accumulate elemental contribution
        
        % mark element for refinement, if still needed
        if max(rho(end)) > rhoMax % refinement criteria
            elementsForRefinement = [elementsForRefinement;iElem];
        end
        
    end
    
    % stop refinement if maximum number of surface elements is reached
    if nOfElements + 3*length(elementsForRefinement) > maxNumberOfElements
        break
    end
end


%% STORE INFORMATION COMPUTED----------------------------------------------
M = zeros(nOfElements,5); % initialise storage matrix
for iElem=1:nOfElements
    
    if elementType == 1
        % vertex coordinates
        v1 = connectivity(1,:,iElem);
        v2 = connectivity(2,:,iElem);
        v3 = connectivity(3,:,iElem);
        % compute unit normal vector at the centroid of each element
        N = (v1+v2+v3)/norm(v1+v2+v3,2);
        plotSphericalTriangle(connectivity(:,:,iElem)) % (uncomment for plotting mesh)
        
    elseif elementType == 2
        % angular coordinates
        theta1=connectivity(iElem,1); theta2=connectivity(iElem,2);
        phi1=connectivity(iElem,3); phi2=connectivity(iElem,4);
        % compute unit normal vector at the centroid of each element
        phi = (phi1 + phi2)/2; theta = acos((cos(theta1)+cos(theta2))/2);
        z = cos(theta); x = sin(theta)*cos(phi); y = sin(theta)*sin(phi);
        N = [x y z];
        %plotSphericalElement(theta1,theta2,phi1,phi2) % (uncomment for plotting mesh)
    end
    M(iElem,:) = [iElem N rho(iElem)];
end
dlmwrite('rho_n.txt',M,'Delimiter',' ','precision',10)
fc_lib_file_add_header_to_file('rho_n','.txt',num2str(nOfElements))
fc_lib_file_add_header_to_file('rho_n','.txt','amesh')
%--------------------------------------------------------------------------
%--------------------------------------------------------------------------

plotElements(connectivity,nOfElements,elementType)

