from math import pi
from scipy.optimize import fsolve
from scipy.interpolate import interp1d

import numpy as np

class layer:
    def __init__(self, name, materials):
        self.name = name
        self.materials = materials
        self.muscle_tone = 0.0
        
#================================ Deformation =================================

    def current_radius(self, R):
        r = np.sqrt( self.r_inner**2 + ( R**2  - self.R_inner**2 )/(self.lmbd_z*self.angle_ratio) )
        return r
    
    def reference_radius(self, r):
        R = np.sqrt( self.R_inner**2 + (self.lmbd_z*self.angle_ratio)*( r**2  - self.r_inner**2 ) )
        return R
    
    def update_current_boundaries(self):
        self.r_inner = self.current_radius(self.R_inner)
        self.r_outer = self.current_radius(self.R_outer)

    def update_reference_boundaries(self):
        self.R_inner = self.reference_radius(self.r_inner)
        self.R_outer = self.reference_radius(self.r_outer)

    def circumferential_stretch(self):
        R = self.reference_radius(self.r)
        lmbd_phi = abs(self.angle_ratio)*self.r/R
        return lmbd_phi

    def circumferential_stretch_inverse(self, lmbd_phi):
        R = np.sqrt(self.angle_ratio*(self.angle_ratio*self.lmbd_z*self.r_inner**2 - self.R_inner**2)/(self.lmbd_z*lmbd_phi**2 - self.angle_ratio))        
        if np.isclose(R[0], self.R_inner):
            R[0] = 1.001*self.R_inner
        if np.isclose(R[-1], self.R_outer):
            R[-1] = 0.999*self.R_outer            
        return R
        
    def deformation_gradient(self, lmbd_phi, gamma):
        try:
            F = np.array([ np.array([1/(lmbd_phi[k]*self.lmbd_z), gamma[k], 0., 0., lmbd_phi[k], 0., 0., 0., self.lmbd_z]).reshape((3, 3)) for k in range(len(lmbd_phi)) ])
        except TypeError:
            F = np.array([ np.array([1/(lmbd_phi   *self.lmbd_z), gamma   , 0., 0., lmbd_phi   , 0., 0., 0., self.lmbd_z]).reshape((3, 3))])
        return F

    def angle_change(self):
        integrand = self.gamma/self.reference_radius(self.r)
        step = self.r[1] - self.r[0]
        integral = abs(step)/2*(2*np.cumsum(integrand) - integrand[0] - integrand)
        self.omega = 180/pi*self.lmbd_z*integral

#============================== Residual Stress ===============================

    def save_residual_stresses(self, R0, Sigma0):
        self.S_res_RR = interp1d(R0, Sigma0[:,0,0], kind='linear')
        self.S_res_PhiPhi = interp1d(R0, Sigma0[:,1,1], kind='linear')
        self.S_res_ZZ = interp1d(R0, Sigma0[:,2,2], kind='linear')

    def residual_stress(self, F):
        lmbd_phi = F[:,1,1]
        R = self.circumferential_stretch_inverse(lmbd_phi)
                  
        S_res = np.array([ np.diag([self.S_res_RR(x), self.S_res_PhiPhi(x), self.S_res_ZZ(x)]) for x in R])
        sigma_res = np.array([ F[k]@S_res[k]@F[k].T for k in range(len(F)) ])
        return sigma_res

#============================ Initial Deformation =============================

    def save_in_vivo_deformation(self, R0, lmbd_phi0, lmbd_z0):
        self.lmbd_phi0 = interp1d(R0, lmbd_phi0, kind='linear')
        self.lmbd_z0 = lmbd_z0

    def in_vivo_deformation_gradient(self, F):
        lmbd_phi = F[:,1,1]
        R = self.circumferential_stretch_inverse(lmbd_phi)
        F0 = [ np.diag([1/(self.lmbd_z0*self.lmbd_phi0(x)), self.lmbd_phi0(x), self.lmbd_z0]) for x in R ]       
        return F0

#================================= Stresses ===================================

    def isochoric_cauchy_stress(self, F, F0=None):
        sigma_iso = np.zeros_like(F)
        for material in self.materials:
            try:
                sigma_iso += material.isochoric_cauchy_stress(F)
            except TypeError:
                try:
                    if F0 is None:
                        F0 = self.in_vivo_deformation_gradient(F)
                        # print(len(F0), len(F))
                    assert len(F) == len(F0), 'The number of material points at which F and F0 is required must be the same'
                    sigma_iso += self.muscle_tone*material.isochoric_cauchy_stress(F, F0)
                except AttributeError:
                    pass
        try:
            sigma_res = self.residual_stress(F)
            sigma_iso += sigma_res
        except AttributeError:
            pass
        return sigma_iso

    def radial_cauchy_stress(self):
        self.r = np.linspace(self.r_inner, self.r_outer, self.n_steps)
        self.lmbd_phi = self.circumferential_stretch()

        def null_shear_stress(x, y):
            gamma, = x
            F = self.deformation_gradient(y, gamma)
            sigma_iso = self.isochoric_cauchy_stress(F)
            return sigma_iso[0,0,1]

        if self.muscle_tone == 0.:
            self.gamma = np.zeros_like(self.r)
        else:
            for material in self.materials:
                try:
                    if np.isclose(material.H[0,1], 0., atol=1e-6) is not True:
                        gamma0 = -0.1*np.sign(material.H[0,1])
                except AttributeError:
                    pass
            self.gamma = [fsolve(null_shear_stress, gamma0, args=y)[0] for y in self.lmbd_phi]
        F = self.deformation_gradient(self.lmbd_phi, self.gamma)
        sigma_iso = self.isochoric_cauchy_stress(F)

        assert np.allclose(sigma_iso[:,0,1], np.zeros_like(self.r), atol=1e-6) is True, 'The circumferential equilibrium equation is not satisfied.'

        integrand = (sigma_iso[:,1,1] - sigma_iso[:,0,0])/self.r
        step = self.r[1] - self.r[0]
        integral = abs(step)/2*(2*np.cumsum(integrand) - integrand[0] - integrand)
        return integral
