from scipy.special import iv, erf
from scipy.linalg import sqrtm, inv
from math import pi
import fitting as fit

import numpy as np

class general_material:
    def __init__(self, name, par_mech):
        self.name = name
        self.par_mech = par_mech

class ground_matrix(general_material):
    def __init__(self, name, par_mech):
        general_material.__init__(self, name, par_mech)
        self.mu_gr = par_mech

    def isochoric_cauchy_stress(self, F):
        sigma_gr = np.array([ self.mu_gr*F[k]@F[k].T for k in range(len(F)) ])
        return sigma_gr

class anisotropic_material(general_material):
    def __init__(self, name, par_mech, par_hist):
        general_material.__init__(self, name, par_mech)
        
        data = fit.read_csv(par_hist, delimiter=" ")
        
        N = data[:,1:4].T
        self.N = N[[2,0,1],:]
        self.rho = data[:,-1]
        self.NxN = np.einsum("ij,kj -> ikj", self.N, self.N)

class collagen(anisotropic_material):
    def __init__(self, name, par_mech, par_hist):
        anisotropic_material.__init__(self, name, par_mech, par_hist)
        self.k1, self.k2 = par_mech

    def isochoric_cauchy_stress(self, F):
        def fibre_contribution(Fi):
            h = ((Fi@self.NxN).T@Fi.T).T
            I = np.trace(h)
            sigma_coli = np.sum(self.rho*(I > 1.0)*2*self.k1*( I - 1.0 )*np.exp( self.k2*( I - 1.0 )**2 )*h, 2)
            return sigma_coli
        sigma_col = np.array(list(map(lambda Fi: fibre_contribution(Fi), F)))
        return sigma_col

class elastin(anisotropic_material):
    def __init__(self, name, par_mech, par_hist):
        anisotropic_material.__init__(self, name, par_mech, par_hist)
        self.c1, self.c2 = par_mech

    def isochoric_cauchy_stress(self, F):
        def fibre_contribution(Fi):
            h = ((Fi@self.NxN).T@Fi.T).T
            I = np.trace(h)
            sigma_eli = np.sum(self.rho*(I > 1.0)*self.c1*( I**(self.c2/2 - 1.0) - 1/I )*h, 2)
            return sigma_eli
        sigma_el = np.array(list(map(lambda Fi: fibre_contribution(Fi), F)))
        return sigma_el

class smooth_muscle(anisotropic_material):
    def __init__(self, name, par_mech, par_hist):
        anisotropic_material.__init__(self, name, par_mech, par_hist)
        self.mu_smc, self.I_opt, self.beta = par_mech


    def isochoric_cauchy_stress(self, F, F0):
        def fibre_contribution(Fi, F0i):
            h = ((Fi@self.NxN).T@Fi.T).T
            I0 = np.sum(F0i@self.N*(F0i@self.N), 0)
            sigma_smci = np.sum(self.rho*self.mu_smc*(I0/self.I_opt)**(self.beta - 1.0)* \
                                np.exp(-(self.beta - 1.0)/self.beta*( (I0/self.I_opt)**self.beta - 1.0) )*h, 2)
            return sigma_smci
        sigma_smc = np.array(list(map(lambda Fi, F0i: fibre_contribution(Fi, F0i), F, F0)))
        return sigma_smc
