from scipy.optimize import fsolve
from scipy.integrate import trapz
from math import pi

import numpy as np
import matplotlib.pyplot as plt

class model:
    def __init__(self, name, layers, geometry, BC):
        self.name = name
        self.n_steps = 300
        self.layers = layers
       
        self.geometry = geometry
        if type(self.geometry['reference']) is tuple and self.geometry['current'] is None:
            self.fixed = 'reference'
            for k, layer in enumerate(self.layers):
                layer.R_inner, layer.R_outer = self.geometry['reference'][k:k+2]
                layer.H = abs(layer.R_outer - layer.R_inner)
        elif type(self.geometry['current']) is tuple and self.geometry['reference'] is None:
            self.fixed = 'current'
            for k, layer in enumerate(self.layers):
                layer.r_inner, layer.r_outer = self.geometry['current'][k:k+2]
                layer.h = abs(layer.r_outer - layer.r_inner)
        else:
            raise Exception('Either the reference or the current configuration must be fixed at the start of the simulation')
        self.BC = BC
        self.lmbd_z, self.opening_angle = BC['essential']
        self.angle_ratio = 180.0/(180.0 - self.opening_angle)
        self.P_inner, self.P_outer, self.F_z = BC['natural']
        self.P_inner *= 0.1333
        self.P_outer *= 0.1333
        
        assert (self.lmbd_z is None or self.F_z is None) and not (self.lmbd_z is None and self.F_z is None), \
                'Either the axial stretch or the axial force must be specified.'
                
        for layer in self.layers:
            layer.lmbd_z = self.lmbd_z
            layer.angle_ratio = self.angle_ratio
            layer.n_steps = round(self.n_steps/len(self.layers))

#=============================== Deformation ==================================

    def generate_current_boundaries(self, a):
        for k, layer in enumerate(self.layers):
            if k == 0:
                layer.r_inner = a
            else:
                layer.r_inner = self.layers[k-1].r_outer
            layer.update_current_boundaries()
            layer.h = abs(layer.r_outer - layer.r_inner)

    def generate_reference_boundaries(self, A):
        for k, layer in enumerate(self.layers):
            if k == 0:
                layer.R_inner = A
            else:
                layer.R_inner = self.layers[k-1].R_outer
            layer.update_reference_boundaries()
            layer.H = abs(layer.R_outer - layer.R_inner)

    def update_geometry(self):
        if self.fixed == 'reference':
            self.geometry['current'] = tuple()
            for layer in self.layers:
                self.geometry['current'] += (layer.r_inner,)
            self.geometry['current'] += (self.layers[-1].r_outer,)
        elif self.fixed == 'current':
            self.geometry['reference'] = tuple()
            for layer in self.layers:
                self.geometry['reference'] += (layer.R_inner,)
            self.geometry['reference'] += (self.layers[-1].R_outer,)

#================================ Stresses ====================================

    def radial_cauchy_stress(self):
        for k, layer in enumerate(self.layers):
            if k == 0:
                sigma_rr = layer.radial_cauchy_stress() - self.P_inner
            else:
                sigma_rr = np.concatenate((sigma_rr, layer.radial_cauchy_stress() + sigma_rr[-1]))
        return sigma_rr

    def axial_force(self, x):
        try:
            self.lmbd_z, = x
        except TypeError:
            self.lmbd_z = x
        for layer in self.layers:
            layer.lmbd_z = self.lmbd_z
        self.admissable_radial_deformation()
        self.cauchy_stress()
        F_z = -0.001*pi*(self.layers[0].r_inner**2*self.P_inner - self.layers[-1].r_outer**2*self.P_outer)
        for layer in self.layers:
            integrand = layer.sigma[:,2,2]*layer.r
            F_z += 0.001*2*pi*trapz(integrand, layer.r)
        return F_z

    def radial_equilibrium(self, x):
        if self.fixed == 'reference':
            a, = x
            self.generate_current_boundaries(a)
        elif self.fixed == 'current':
            A, = x
            self.generate_reference_boundaries(A)
        self.sigma_rr = self.radial_cauchy_stress()
        root = self.sigma_rr[-1] + self.P_outer
        
        return root

    def axial_equilibrium(self, x):
        F_z = self.axial_force(x)
        root = F_z - self.F_z
        return root

    def cauchy_stress(self):
        boundaries = (0,) + tuple(k+1 for k in range(len(self.sigma_rr)-1) if self.sigma_rr[k] == self.sigma_rr[k+1]) + (len(self.sigma_rr)+1,)
        
        for k, layer in enumerate(self.layers):
            start = boundaries[k]
            end = boundaries[k+1]
            sigma_rr = self.sigma_rr[start:end]

            F = layer.deformation_gradient(layer.lmbd_phi, layer.gamma)
            sigma_iso = layer.isochoric_cauchy_stress(F)

            p = sigma_iso[:,0,0] - sigma_rr

            layer.sigma = np.array([ sigma_iso[k,:,:] - p[k]*np.eye(3) for k in range(len(F)) ])

#========================== Boundary Value Problem ============================

    def admissable_radial_deformation(self):
        if self.fixed == 'reference':
            x0 = 1/abs(self.angle_ratio)*self.layers[-1].R_inner
        elif self.fixed == 'current':
            x0 = abs(self.angle_ratio)*self.layers[-1].r_inner   
            
        x = fsolve(self.radial_equilibrium, x0)  
        
        if self.fixed == 'reference':
            a, = x
            self.generate_current_boundaries(a)
        elif self.fixed == 'current':
            A, = x
            self.generate_reference_boundaries(A)
            
        assert np.isclose(self.sigma_rr[-1], self.P_outer), 'The radial equilibrium equation is not satisfied.'

    def admissable_axial_deformation(self):
        self.lmbd_z, = fsolve(self.axial_equilibrium, 1.)
        for layer in self.layers:
            layer.lmbd_z = self.lmbd_z

    def BVP_solver(self):
        if self.lmbd_z is None:
            self.admissable_axial_deformation()
            self.admissable_radial_deformation()
        elif self.F_z is None:
            self.F_z = self.axial_force(self.lmbd_z)
        for k, layer in enumerate(self.layers):
            layer.angle_change()
            if k > 0:
                layer.omega += self.layers[k-1].omega[-1]       
                
        self.update_geometry()
        
        if self.fixed == 'reference' and self.angle_ratio == 1.0:
            self.h = self.geometry['current'][-1] - self.geometry['current'][0]
            self.H = self.geometry['reference'][-1] - self.geometry['reference'][0]
            self.lmbd_phi = self.H/(self.lmbd_z*self.h)
            self.gamma = 0.
            for layer in self.layers[:-1]:
                self.gamma += np.trapz(layer.gamma, layer.r)
            self.gamma /= self.layers[-2].r_outer - self.layers[0].r_inner
            self.omega = self.layers[-1].omega[-1]
            
        self.cauchy_stress()

#============================== Graphical Output ==============================

    def plot_results(self):
        plt.rcParams.update({
            "axes.axisbelow": True,
            "font.family": "serif",
            "font.size": 9,
            "font.serif": [],
            "font.sans-serif": [],
            "text.usetex": True,
            "text.latex.preamble": [r"\usepackage{upgreek}", r"\usepackage{amsmath}"]
            })
    
        _, ax_deformation = plt.subplots(2, 1, sharex=True, figsize=tuple(1*14*0.39370*np.array([1, 1])))
        ax_deformation[0].set_ylabel(r'Circumferential Stretch $\lambda_\varphi$ / 1')
        ax_deformation[0].grid()
        ax_deformation[1].set_ylabel(r'Shear Deformation $\omega$ / ${}^\circ$')
        ax_deformation[1].grid()
        ax_deformation[-1].set_xlabel(r'Current Radius r / mm')
        ax_deformation[-1].set_xlim(self.geometry['current'][0], self.geometry['current'][-1])         
        
        for layer in self.layers:
            ax_deformation[0].plot(layer.r, layer.lmbd_phi, lw=1.)
            ax_deformation[1].plot(layer.r, layer.omega, lw=1.)

        plt.tight_layout()
        plt.show()

        _, ax_stress = plt.subplots(3, 1, sharex=True, figsize=tuple(1*14*0.39370*np.array([1, 1.5])))
        ax_stress[-1].set_xlim(self.geometry['current'][0], self.geometry['current'][-1])      
        ax_stress[-1].set_xlabel(r'Current Radius r / mm')

        for layer in self.layers:
            for k, direction in enumerate((r'{rr}', r'{\varphi\varphi}', r'{zz}')):
                ax_stress[k].grid()
                ax_stress[k].plot(layer.r, layer.sigma[:,k,k], lw=1.)
                ax_stress[k].set_ylabel(r'\textsc{Cauchy} Stress $\upsigma_'+direction+'$ / kPa')
                
        plt.tight_layout()
        plt.show()

    
    










